/*------------------------------------------------------------------------------
* convrnx.c : convert receiver raw/rtcm data to rinex and sbas log file
*
*          Copyright (C) 2009 by T.TAKASU, All rights reserved.
*
* version : $Revision: 1.1 $ $Date: 2008/07/17 21:48:06 $
* history : 2009/04/10 1.0 new
*           2009/06/02 1.1 support glonass
*           2009/12/18 1.2 add check return of init_rtcm()/init_raw()
*-----------------------------------------------------------------------------*/
#include "rtklib.h"

static const char rcsid[]="$Id:$";

/* application defined function to output message ----------------------------*/
extern int showmsg(char *format,...);

/* save slip conditions ------------------------------------------------------*/
static void saveslips(unsigned char slips[][NFREQ], obsd_t *data, int n)
{
    int i,j;
    
    for (i=0;i<n;i++) for (j=0;j<NFREQ;j++) {
        if (data[i].LLI[j]&1) slips[data[i].sat-1][j]|=1;
    }
}
/* restore slip conditions ---------------------------------------------------*/
static void restslips(unsigned char slips[][NFREQ], obsd_t *data, int n)
{
    int i,j;
    
    for (i=0;i<n;i++) for (j=0;j<NFREQ;j++) {
        if (slips[data[i].sat-1][j]&1) data[i].LLI[j]|=1;
        slips[data[i].sat-1][j]=0;
    }
}
/* rinex converter -------------------------------------------------------------
* convert receiver log file to rinex obs/nav, sbas log files
* args   : int    format I      receiver raw format (STRFMT_???)
*          rnxopt_t *opt I      rinex options
*          char   *file  I      receiver raw file
*          char   *ofile I      output rinex obs file  ("": no output)
*          char   *nfile I      output rinex nav file  ("": no output)
*          char   *gfile I      output rinex gnav file ("": no output)
*          char   *sfile I      output sbas log file   ("": no output)
* return : status (1:ok,0:error)
*-----------------------------------------------------------------------------*/
extern int convrnx(int format, const rnxopt_t *opt, const char *file, 
                   const char *ofile, const char *nfile, const char *gfile,
                   const char *sfile)
{
    FILE *fp,*ofp=NULL,*nfp=NULL,*gfp=NULL,*sfp=NULL;
    rnxopt_t rnxopt={{0}};
    raw_t  raw ={{0}};
    rtcm_t rtcm={0};
    obs_t *obs;
    nav_t *nav;
    gtime_t time={0},ts={0},tstart,tend;
    int i,ret,sys,sat,prn,n[5]={0};
    char s[64]="";
    unsigned char slips[MAXSAT][NFREQ]={{0}};
    
    trace(3,"convrnx: format=%d file=%s ofile=%s nfile=%s sfile=%s\n",
          format,file,ofile,nfile,sfile);
    
    rnxopt=*opt; showmsg("");
    
    if (format==STRFMT_RTCM2||format==STRFMT_RTCM3) {
        if (!init_rtcm(&rtcm)) {
            showmsg("init_raw error");
            return 0;
        }
        if (!rnxopt.trtcm.time) {
            showmsg("Needs approximate Start Time for RTCM 2/3");
            return 0;
        }
        rtcm.time=rnxopt.trtcm; /* initial time */
        obs=&rtcm.obs; nav=&rtcm.nav; 
        strcpy(rtcm.opt,rnxopt.rcvopt);
    }
    else {
        if (!init_raw(&raw)) {
            showmsg("init_raw error");
            return 0;
        }
        obs=&raw.obs; nav=&raw.nav;
        strcpy(raw.opt,rnxopt.rcvopt);
    }
    if (!(fp=fopen(file,"rb"))) {
        showmsg("Log open error: %s",file);
        return 0;
    }
    if (*ofile&&(!strcmp(ofile,file)||!(ofp=fopen(ofile,"w")))) {
        showmsg("OBS open error: %s",ofile);
        fclose(fp);
        return 0;
    }
    if (*nfile&&(!strcmp(nfile,file)||!(nfp=fopen(nfile,"w")))) {
        showmsg("NAV open error: %s",nfile);
        fclose(fp); fclose(ofp);
        return 0;
    }
    if (*gfile&&(!strcmp(gfile,file)||!(gfp=fopen(gfile,"w")))) {
        showmsg("GNAV open error: %s",gfile);
        fclose(fp); fclose(ofp); fclose(nfp);
        return 0;
    }
    if (*sfile&&(!strcmp(sfile,file)||!(sfp=fopen(sfile,"w")))) {
        showmsg("SBAS Log open error: %s",sfile);
        fclose(fp); fclose(ofp); fclose(gfp); fclose(nfp);
        return 0;
    }
    for (i=0;;i++) {
        if (format==STRFMT_RTCM2) {
            if ((ret=input_rtcm2f(&rtcm,fp))<-1) break;
            time=rtcm.time; sat=rtcm.ephsat;
        }
        else if (format==STRFMT_RTCM3) {
            if ((ret=input_rtcm3f(&rtcm,fp))<-1) break;
            time=rtcm.time; sat=rtcm.ephsat;
        }
        else {
            if ((ret=input_rawf(&raw,format,fp))<-1) break;
            time=raw.time; sat=raw.ephsat;
        }
        if (ofp&&ret==1) {
            saveslips(slips,obs->data,obs->n); /* save slip conditions */
            if (screent(time,rnxopt.ts,rnxopt.te,rnxopt.tint)) {
                if (n[0]==0) {
                    rnxopt.tstart=obs->data[0].time;
                    outrnxobsh(ofp,&rnxopt);
                }
                restslips(slips,obs->data,obs->n); /* restore slip conditions */
                outrnxobsb(ofp,opt,obs->data,obs->n,0);
                n[0]++;
            }
        }
        else if (nfp&&ret==2) {
            tstart=rnxopt.ts;
            tend  =rnxopt.te;
            if (tstart.time!=0) tstart=timeadd(tstart,-MAXDTOE);
            if (tend  .time!=0) tend  =timeadd(tend  , MAXDTOE);
            if (screent(time,tstart,tend,0.0)) {
                if (n[1]==0) outrnxnavh(nfp,&rnxopt,nav->ion,nav->utc,nav->leaps);
                outrnxnavb(nfp,&rnxopt,nav->eph+sat-1);
                n[1]++;
             }
        }
        else if (gfp&&ret==4) {
            sys=satsys(sat,&prn);
            tstart=rnxopt.ts;
            tend  =rnxopt.te;
            if (tstart.time!=0) tstart=timeadd(tstart,-MAXDTOE_GLO);
            if (tend  .time!=0) tend  =timeadd(tend  , MAXDTOE_GLO);
            if (sys==SYS_GLO&&screent(time,tstart,tend,0.0)) {
                if (n[2]==0) outrnxgnavh(gfp,&rnxopt);
                outrnxgnavb(gfp,&rnxopt,nav->geph+prn-1);
                n[2]++;
             }
        }
        else if (sfp&&ret==3) {
            if (screent(time,rnxopt.ts,rnxopt.te,0.0)) {
                sbsoutmsg(sfp,&raw.sbsmsg);
                n[3]++;
            }
        }
        else if (ret==-1) n[4]++;
        
        if (ts.time==0) ts=time;
        if (rnxopt.te.time&&timediff(time,rnxopt.te)>-DTTOL) break;
        if (i%11) continue;
        if (time.time!=0) {
            time2str(time,s,0); strcat(s,": ");
        }
        if (showmsg("%sO=%d N=%d G=%d S=%d E=%d",s,n[0],n[1],n[2],n[3],n[4])) {
            break;
        }
    }
    if (format==STRFMT_RTCM2||format==STRFMT_RTCM3) {
        sprintf(rnxopt.comment[1]+strlen(rnxopt.comment[1]),
                ", station ID: %d",rtcm.staid);
        for (i=0;i<3;i++) rnxopt.apppos[i]=rtcm.sta.rr[i]+rtcm.sta.del[0][i];
        rnxopt.antdel[2]=rtcm.sta.anth;
    }
    if (ofp) {
        rewind(ofp);
        outrnxobsh(ofp,&rnxopt);
        fclose(ofp);
    }
    if (nfp) {
        rewind(nfp);
        outrnxnavh(nfp,&rnxopt,nav->ion,nav->utc,nav->leaps);
        fclose(nfp);
    }
    if (gfp) {
        rewind(gfp);
        outrnxgnavh(gfp,&rnxopt);
        fclose(gfp);
    }
    if (sfp) {
        fclose(sfp);
    }
    fclose(fp);
    
    if (ts.time!=0) {
        time2str(time,s+15,0); time2str(ts,s,0); s[19]='-'; strcat(s,": ");
    }
    showmsg("%sO=%d N=%d G=%d S=%d E=%d",s,n[0],n[1],n[2],n[3],n[4]);
    
    free_rtcm(&rtcm); free_raw(&raw);
    return 1;
}
